from importlib.metadata import version as _version

from fast_vindex.api import patched_vindex

try:
    __version__ = _version("fast_vindex")
except Exception:
    __version__ = "9999"

__all__ = ["patched_vindex"]



from itertools import product

import dask.array as da
import numpy as np


def _narray_indices(shape):
    """Generate a NumPy array containing the index values."""
    narray = np.empty(shape, dtype=object)
    for indices in product(*(range(s) for s in shape)):
        narray[indices] = str(indices)
    return narray


# def _narray_random(shape, seed=None):
#     """Generate a NumPy array with random values."""
#     rs = np.random.default_rng(seed)
#     return rs.random(shape)


def _darray_indices(shape, chunks=None):
    """Generate a Dask array containing the index values."""
    narray = _narray_indices(shape)
    darray = da.from_array(narray, chunks=chunks)
    return darray


def _darray_random(shape, chunks=None, seed=None):
    """Generate a Dask array with random values."""
    rs = da.random.RandomState(seed=seed)
    return rs.random(shape, chunks=chunks)


def generate_darray(shape, chunks=None, fmt="drandom", seed=None):
    """Generate a dask array of the requested type and format."""
    if fmt == "drandom":
        return _darray_random(shape, chunks, seed)
    elif fmt == "dindices":
        return _darray_indices(shape, chunks)
    else:
        raise ValueError(
            f"Format '{fmt}' not recognized. Use 'drandom', or 'dindices'."
        )


def generate_fancy_indexes(darray: da.Array, n: int, padding: int = 0):
    """Generate a set of random indexes for a given Dask array"""
    offsets = np.arange(-padding, padding)
    indexes = []
    for axis, size in enumerate(darray.shape):
        low, high = padding, size - padding
        base_indices = np.random.randint(low, high, size=(n, 1))
        expanded = base_indices + offsets
        shape = [1] * darray.ndim
        shape[axis] = 2 * padding
        expanded = expanded.reshape((n, *shape))
        indexes.append(expanded)
    return tuple(indexes)


def estimate_graph_size(collections):
    from collections.abc import Iterator

    import dask
    from dask.base import collections_to_expr
    from dask.utils import format_bytes
    from distributed.protocol import serialize, to_serialize
    from distributed.protocol.serialize import Serialized
    from distributed.utils import nbytes

    # issue de distributed/client.py/compute()
    collections = tuple(
        (dask.delayed(a) if isinstance(a, (list, set, tuple, dict, Iterator)) else a)
        for a in collections
    )
    variables = [a for a in collections if dask.is_dask_collection(a)]

    kwargs = {}
    optimize_graph = True
    expr = collections_to_expr(variables, optimize_graph, **kwargs)

    # issue de distributed/client.py/_graph_to_futures()
    expr_ser = Serialized(*serialize(to_serialize(expr), on_error="raise"))
    pickled_size = sum(map(nbytes, [expr_ser.header] + expr_ser.frames))
    return pickled_size, format_bytes(pickled_size)



from fast_vindex.testing.core import generate_darray, generate_fancy_indexes
from fast_vindex.testing.plot import Array2D, DaskArray2D

__all__ = [
    "generate_darray",
    "generate_fancy_indexes",
    "Array2D",
    "DaskArray2D",
]



import dask.array as da
import matplotlib.pyplot as plt
import numpy as np


class Array2D:
    """
    Surcouche Numpy pour représenter la sélection d'une zone dans un array 2D.

    arr = np.arange(81).reshape(9, 9)
    a2d = Array2D(arr)
    a2d[indexes]
    """

    def __init__(self, array2d):
        assert array2d.ndim == 2, "L'array doit être 2D"
        self.array = array2d
        self.dims = array2d.shape

    def __getitem__(self, key):
        selection = np.zeros(self.dims, dtype=bool)
        selection[key] = True

        fig, ax = plt.subplots()
        ax.imshow(np.ones(self.dims), cmap="Greys", alpha=0.1)  # fond gris clair
        ax.imshow(selection, cmap="Blues", alpha=0.6)

        # Ajout de la grille (lignes fines entre les cases)
        ax.set_xticks(np.arange(-0.5, self.dims[1], 1), minor=True)
        ax.set_yticks(np.arange(-0.5, self.dims[0], 1), minor=True)
        ax.grid(which="minor", color="black", linestyle="-", linewidth=0.5)

        # Afficher les ticks principaux avec les indices
        ax.set_xticks(np.arange(self.dims[1]))
        ax.set_yticks(np.arange(self.dims[0]))
        ax.set_xticklabels(np.arange(self.dims[1]))
        ax.set_yticklabels(np.arange(self.dims[0]))

        # ax.invert_yaxis()
        ax.xaxis.set_ticks_position("top")
        ax.xaxis.set_label_position("top")
        plt.show()

    def __repr__(self):
        return self.array.__repr__()


class DaskArray2D:
    """
    arr = da.arange(81).reshape(9, 9).rechunk((3, 3))
    da2d = DaskArray2D(arr)
    da2d[indexes]
    """

    def __init__(self, dask_array):
        if not isinstance(dask_array, da.Array):
            raise TypeError("L'argument doit être un dask.array.Array")
        if dask_array.ndim != 2:
            raise ValueError("DaskArray2D ne supporte que les arrays 2D")

        self.arr = dask_array
        self.dims = dask_array.shape
        self.chunk_sizes = dask_array.chunks  # tuple de tuples (par dimension)

    def __getitem__(self, key):
        selection = np.zeros(self.dims, dtype=bool)
        selection[key] = True

        fig, ax = plt.subplots()
        ax.imshow(np.ones(self.dims), cmap="Greys", alpha=0.1)
        ax.imshow(selection, cmap="Blues", alpha=0.6)

        # Ticks majeurs centrés sur les pixels (indices)
        ax.set_xticks(np.arange(self.dims[1]), minor=False)
        ax.set_yticks(np.arange(self.dims[0]), minor=False)

        # Labels pour ticks
        ax.set_xticklabels(np.arange(self.dims[1]))
        ax.set_yticklabels(np.arange(self.dims[0]))

        # Grille mineure entre pixels
        ax.set_xticks(np.arange(-0.5, self.dims[1], 1), minor=True)
        ax.set_yticks(np.arange(-0.5, self.dims[0], 1), minor=True)
        ax.grid(which="minor", color="black", linestyle="-", linewidth=0.5)

        # Fonction utilitaire pour positions cumulées des chunks
        def cumsum_chunks(chunks):
            positions = [0]
            for size in chunks:
                positions.append(positions[-1] + size)
            return positions

        # Positions lignes chunks (à -0.5 pour coller à la grille)
        x_chunk_lines = cumsum_chunks(self.chunk_sizes[1])
        y_chunk_lines = cumsum_chunks(self.chunk_sizes[0])
        x_chunk_lines_shifted = [x - 0.5 for x in x_chunk_lines]
        y_chunk_lines_shifted = [y - 0.5 for y in y_chunk_lines]

        # Tracer lignes de séparation des chunks avec axvline et axhline
        for x in x_chunk_lines_shifted:
            ax.axvline(x=x, color="red", linewidth=2)
        for y in y_chunk_lines_shifted:
            ax.axhline(y=y, color="red", linewidth=2)

        # ax.invert_yaxis()
        ax.xaxis.set_ticks_position("top")
        ax.xaxis.set_label_position("top")
        plt.show()
